#!/usr/bin/perl

# This script generates an HTML with coreference outputs.
# Also GOLD is generated if it exists.

use FindBin qw( $RealBin );
use lib "$RealBin/lib";

use strict;
use Getopt::Long;
use Document qw (%FIXCOL);
use CorScorer;

###############################################################################
# globals

use Globals qw ($PATH_HOME $PATH_MODELS $STEM);

my $VERBOSE = 0;

my $DEFAULT_SECTION = 'sec-default';
my $DEFAULT_MODEL = 'default';

my @HTMLcolors = ('#F0F8FF', '#FAEBD7', '#00FFFF', '#7FFFD4', '#F5F5DC',
						'#000000', '#0000FF', '#8A2BE2', '#A52A2A', '#DEB887', '#5F9EA0',
						'#7FFF00', '#FF7F50', '#6495ED', '#FFF8DC', '#DC143C', 
						'#00008B', '#008B8B', '#B8860B', '#A9A9A9', '#006400', '#BDB76B', '#8B008B', 
						'#556B2F', '#FF8C00', '#9932CC', '#8B0000', '#E9967A', '#8FBC8F', '#483D8B',
						'#2F4F4F', '#00CED1', '#9400D3', '#FF1493', '#00BFFF', '#696969', '#1E90FF',
						'#B22222', '#FFFAF0', '#228B22', '#FF00FF', '#DCDCDC', '#F8F8FF', '#FFD700',
						'#DAA520', '#808080', '#008000', '#ADFF2F', '#F0FFF0', '#FF69B4', '#CD5C5C',
						'#4B0082', '#FFFFF0', '#F0E68C', '#E6E6FA', '#FFF0F5', '#7CFC00', '#FFFACD',
						'#ADD8E6', '#F08080', '#E0FFFF', '#FAFAD2', '#D3D3D3', '#90EE90', '#FFB6C1',
						'#FFA07A', '#20B2AA', '#87CEFA', '#778899', '#B0C4DE', '#FFFFE0', '#00FF00',
						'#32CD32', '#FAF0E6', '#FF00FF', '#800000', '#66CDAA', '#0000CD', '#BA55D3',
						'#9370D8', '#3CB371', '#7B68EE', '#00FA9A', '#48D1CC', '#C71585', '#191970',
						'#F5FFFA', '#FFE4E1', '#FFE4B5', '#FFDEAD', '#000080', '#FDF5E6', '#808000',
						'#6B8E23', '#FFA500', '#FF4500', '#DA70D6', '#EEE8AA', '#98FB98', '#AFEEEE',
						'#D87093', '#FFEFD5', '#FFDAB9', '#CD853F', '#FFC0CB', '#DDA0DD', '#B0E0E6',
						'#800080', '#FF0000', '#BC8F8F', '#4169E1', '#8B4513', '#FA8072', '#F4A460',
						'#2E8B57', '#FFF5EE', '#A0522D', '#C0C0C0', '#87CEEB', '#6A5ACD', '#708090',
						'#FFFAFA', '#00FF7F', '#4682B4', '#D2B48C', '#008080', '#D8BFD8', '#FF6347',
						'#40E0D0', '#EE82EE', '#F5DEB3', '#FFFFFF', '#F5F5F5', '#FFFF00', '#9ACD32');
						
my %opos = ('0' => 'F', '1' => 'E', '2' => 'D', '3' => 'C', '4' => 'B', '5' => 'A',
				'6' => '9', '7' => '8', '8' => '7', '9' => '6', 'A' => '5', 'B' => '4',
				'C' => '3', 'D' => '2', 'E' => '1', 'F' => '0');


###############################################################################
# functions


# Primers tags html (html, head, body...)
sub HeadsHtml
{
   my ($title, $file) = @_;
   
   print "<html><head><title>$title | $file</title>\n";
	print "<style type=\"text/css\">\n";
	print '
	body {
		background: #fff;
		font-size: 14px;
	}
	
	#loading {
		position:absolute;
		left: 150px;
		/*top: 100px;*/
		border: none;
		text-align: justify;
		align: center;
		visibility: hidden;
	}
	
	#panel1 {
		position:absolute;
		left: 10px;
		/*top: 100px;*/
		border: #000 1px solid;
		width: 450px;
		text-align: justify;
	}
	
	#panel2 {
		position:absolute;
		left: 470px;
		/*top: 100px;*/
		border: #000 1px solid;
		width: 450px;
		text-align: justify;
	}
	
	';
	print "</style>\n";
	print "</head>\n";
	print "<body>\n";
	print "<h2>$file</h2>\n";
}

# Ultims tags html (html, head, body...)
sub FiHtml
{
   print "</body>\n";
   print "</html>\n";
}


sub Index
{
	my ($files) = @_;
	
	print "<p>Files:</p>\n";
	print "<ul>\n";
	foreach my $file (@$files) {
		my @dirs = split(/\//, $file);
		print "<li><a href=\"$dirs[-1].html\">$file</a></li>\n";
	}
	print "</ul>\n";
}

# no es exactament l'oposat pero serveix
sub ColorOposat
{
	my $color = shift;
	
	my ($red, $green, $blue) = ($color =~ /\#(\w\w)(\w\w)(\w\w)/);
	
	my $fosc = 0;
	foreach my $color ($red, $green, $blue) {
		# Simple, si es clar ho fem negre i si es fosc ho fem blanc
		if ($color =~ /^[0-8]/) {
			$fosc++;
		}
	}
	
	my $newColor;
	if ($fosc >= 2) {
		$newColor = '#ffffff';
	}
	else {
		$newColor = '#000000';
	}
	
	return $newColor;
}

sub EstilChain
{
	my ($chain, $singletons, $ceaf) = @_;
	
	my $color;
	if (defined($singletons->{$chain})) {
		$color = '#e0e0e0';
	}
	else {
		$color = $HTMLcolors[$chain % scalar(@HTMLcolors)];
	}
	
	my $str = "background:$color; color:" . ColorOposat($color) . ";";
	if ($ceaf) {
		$str .= " border-style: solid; border-width: " . (1 + abs($ceaf)) . "px;";
		if ($ceaf > 0) {
			$str .= " border-color: green;";
		}
		else {
			$str .= " border-color: red;";
		}
	}

	return $str;
}


sub ShowFileChains
{
	my ($file, $gold) = @_;
	my %singletons;
	my %id;
	my %ini;
	my %fi;
	
	my $entities;
	if ($gold) {
		$entities = &CorScorer::GetCoreference("$file.gold", $FIXCOL{coref}, '__single');
	}
	else {
		$entities = &CorScorer::GetCoreference("$file.out", $FIXCOL{response}, '__single');
	}
	
	my $idCount = 0;
	for (my $i = 0; $i < @$entities; $i++) {
		next if (!defined($entities->[$i]));
		my $entity = $entities->[$i];
		if (scalar(@$entity) == 1) {
			$singletons{$i} = 1;
		}
		
		foreach my $mention (@$entity) {
			my %m = (long => $mention->[1] - $mention->[0], id => $idCount, entity => $i);
			push(@{$ini{$mention->[0]}}, \%m);
			$fi{$mention->[1]}++;
			$idCount++;
		}
	}
	
	if (open (F, $file)) {
		my $dins = 0;
		my $tcount = 0;
		while (my $l = <F>) {
			chomp($l);
			if ($l eq '') {
				print "</p>\n<p>";
			}
			else {
				my @cols = split(/\t/, $l);
				
				if (defined($ini{$tcount})) {
					foreach my $m (sort {$b->{long} <=> $a->{long}} (@{$ini{$tcount}})) {
						print "<span title=\"$m->{id}\" style=\"";
						print EstilChain($m->{entity}, \%singletons) . "\">";
						if (!defined($singletons{$m->{entity}})) {
							print "<sup>$m->{entity}</sup>";
						}
						print "\[";
					}
				}
				
				print $cols[$FIXCOL{token}];
				
				if (defined($fi{$tcount})) {
					for (my $j = 0; $j < $fi{$tcount}; $j++) {
						print "\]</span>";
					}
				}
				
				print ' ';
				$tcount++;
			}
		}
		close (F);
		print "</p>\n";
	}
	
	if (!$gold) {
		print "\n";
	
		print "<table>";
		print "<tr><td>M</td><td>Rec</td><td>Pre</td><td>F1</td></tr>\n";
		foreach my $m ('ceafm', 'muc', 'bcub', 'ceafe') {
			print "<tr><td>$m</td>";
			my $ooo = select(STDERR);
			my ($numrec, $denrec, $numpre, $denpre) = &CorScorer::Score( $m, "$file.gold",
																							 "$file.out", '__single' );
			select($ooo);
			my $precisio = $denpre ? $numpre / $denpre : 0;
			my $recall = $denrec ? $numrec / $denrec : 0;
			my $f1 = 0;
			if ($recall + $precisio) {
				$f1 = 2 * $precisio * $recall / ($precisio + $recall);
			}
			print "<td>" . int($recall*10000)/100 . '%' . "</td>"; #" ($numrec/$denrec)</td>";
			print "<td>" . int($precisio*10000)/100 . '%' . "</td>"; #" ($numpre/$denpre)</td>";
			print "<td>" . int($f1*10000)/100 . "\%</td>";
			print "</tr>\n";
		}
		print "</table>\n";
	}
}


sub Panel
{
	my ($file, $gold) = @_;
	
	print "<p>$file";
	if ($gold) { print " (gold)"; }
	else { print " (response)"; }
	print "</p>\n";
	ShowFileChains($file, $gold);
}

sub Main
{
	my ($file) = @_;
	
# 	print "<p><a href=\"index.html\">Change file</a></td></tr>\n";
	
	print "<div id=\"panel1\">\n";
	Panel($file, 0);
	print "</div>\n";
	
	print "<div id=\"panel2\">\n";
	Panel($file, 1);
	print "</div>\n";

}

sub GenHTMLFile
{
	my ($file) = @_;
	open (HTML, ">$file.html") || die "Can not write $file.html";
	my $def = select(HTML);
	HeadsHtml("Error analysis", $file);
	Main($file);
	FiHtml;
	select($def);
}

sub GenHTMLIndex
{
	my ($files) = @_;
	
	my ($dir) = ($files->[0] =~ /^(.*?)[^\/]*$/);
	
	open (HTML, ">$dir/index.html") || die "Can not write $dir/index.html";
	my $def = select(HTML);
	HeadsHtml("Error analysis", "Directory index");
	Index($files);
	FiHtml;
	select($def);
}

##############################################################################
# main

my $result;
my %opt;
if (@ARGV) {
	$result = GetOptions (\%opt, 'index', 'help', 'output');
}

my @files = @ARGV;

if (@files == 0 || $opt{'help'}) {
	if (@files == 0) {
		print 'Error: no files in the argument

';
	}
	
	print 'use: ./gen-html.pl [options] <file(s)>
	
		Generates an HTML file showing the coreference chains of response and key
		for qualitative evaluation and error analysis.
		
		Options:
		
			index    Generates an index (index.html)
			
			help     Shows this help message

';
# not implemented yet
# output   Directory to store the output files. If not included, the html
# 			         files are saved in the same directory as files are found.
	exit;
}


if (defined($opt{'index'})) {
	print "Generating index.html...\n";
	GenHTMLIndex(\@files);
}


foreach my $file (@files) {
	print "Generating $file.html...\n";
	GenHTMLFile($file);
}
