package DT;
require Model;
our @ISA = qw(Model);
require Attributes;
use strict;

##############################################################################
# globals

use Globals qw($PATH_C45 $PATH_C45RULES $SAMPLE_NAMES);


##############################################################################
# functions

# This function assumes that all the features are binary
sub GenNames
{
	my ($obj) = @_;
	my $fTree = $obj->modelFile;
	my @features = $obj->features;
	
	open (NAM, ">$fTree.names") || die "No puc obrir $fTree.names";
	print NAM '+, -.   | classes' . "\n\n";
	foreach my $f (@features) {
		print NAM "$f:\t0,1.\n";
	}
	close (NAM);
}

sub GenDT
{
	my ($dt, $opt) = @_;
	
	# Generates the necessary info file for C4.5 and the order
	# of the attributes
	$dt->GenNames unless (defined($opt->{noGenNames}));
	
	my $fTree = $dt->modelFile;
	system("$PATH_C45 $opt->{c45} -f$fTree > $fTree.c45output");
	$dt->ReadDTC45("$fTree.c45output");
}

sub GetTreeRules {
	my ($dt) = @_;
	return $dt->{_treeRules} if (defined($dt->{_treeRules}));
	
	my $fTree = $dt->modelFile;
	my @condicions = ();
	open (RUL, "$fTree.perl") || die "No puc obrir $fTree.perl";
	while (my $l = <RUL>) {
		chomp($l);
		my ($id, $regla, $salta_ok, $salta_ko, $result, $prob) = split(/\|/, $l);
		$condicions[$id]{condicio} = $regla;
		if ($salta_ok == -1) {
			$condicions[$id]{final} = 1;
			$condicions[$id]{result} = $result;
			$condicions[$id]{prob} = $prob;
		}
		else {
			$condicions[$id]{final} = 0;
			$condicions[$id]{seguent_ok} = $salta_ok;
		}
		$condicions[$id]{seguent_ko} = $salta_ko;
	}
	close (RUL);
	
	$dt->{_treeRules} = \@condicions;
	
	return $dt->{_treeRules};
}

sub MyConsult
{
	my ($dt, $ma, $mb) = @_;

	# GenPairFeatures returns a reference to a hash in scalar context
	my $r = $dt->GenPairFeatures($ma, $mb);
	my %a = %{$r};
	my $condicions = $dt->GetTreeRules;
	my $torna;
	my $id = 0;
	my $voltes = 0;
	do {
		$torna = 0;
		$voltes++;
		my $r;
		my $cond = $condicions->[$id]{condicio};
		eval $cond;
		print "$id) $cond ==> r: $r\n" if ($voltes > scalar(@$condicions));
		if ($r) {
			if ($condicions->[$id]{final}) {
				my $myr = ($condicions->[$id]{result} eq '+') ? 0 : 1;
				return ( $myr, $condicions->[$id]{prob});
			}
			else {
				$id = $condicions->[$id]{seguent_ok};
				$torna = 1;
			}
		}
		else {
			$id = $condicions->[$id]{seguent_ko};
			$torna = 1;
		}
	} while ($torna && $id != -1);
	
	# aqui no hauria d'arribar mai
	foreach my $key (keys(%a)) {
		print "$key: $a{$key}\n";
	}
	die "ooh";
	return (1, 0.5);
}


sub ReadDTC45
{
	my ($dt, $fSortida) = @_;
	my $fArbre = $dt->modelFile;
	my $dins = 0;
	my $tincarbre = 0;
	my $id = 0;
	my @condicions;
	my @pendents;
	my $acumnivell = 0;
	open(F, $fSortida) || die "No puc obrir $fSortida";
	while (my $l = <F>) {
		chomp($l);
		if ($dins) {
			if ($l =~ /^([\|\s]*?)(\w+) ([\<\>\=]+) (\w+)\s?:(.*?)$/) {
				my $nivell = $1;
				my $atribut = $2;
				my $condicio = $3;
				my $valor = $4;
				my $result = $5;
				my $prob = 0;
				if ($nivell eq '') {
					$nivell = $acumnivell;
				}
				else {
					$nivell = scalar( split(/\|/, $nivell) ) - 1 + $acumnivell;
				}
				
				if ($condicio eq '=') {
					if ($valor =~ /^\d+$/) {
						$condicio = '==';
					}
					else {
						$condicio = 'eq';
						$valor = "\'$valor\'";
					}
				}
				
				my %cond;
				my $evalua = "\$r = \$a\{\'$atribut\'\} $condicio $valor";
				$cond{condicio} = $evalua;
				$cond{nivell} = $nivell;
				$cond{result} = -1;
				$cond{prob} = -1;
				
				# busquem antecedent entre les pendents
				if ($id > 0 && $nivell < $condicions[$id-1]->{nivell} ) {
					my $trobat = 0;
					do {
						my $idp = pop(@pendents);
						if ($condicions[$idp]->{nivell} == $nivell) {
							$condicions[$idp]->{seguent_ko} = $id;
							$trobat = 1;
						}
					} while (!$trobat && @pendents);
				}
				
				if ($result =~ /^ ([+-])(.*?)$/) {
					$result = $1;
					$prob = $2;
					if ($prob =~ /\(([\d\.]+)\/([\d\.]+)\)/) {
						$prob = $1 / ($1 + $2);
					}
					elsif ($prob =~ /\(([\d\.]+)\)/) {
						if ($1 > 0) {
							$prob = 1;
						}
						else {
							$prob = 0.5;
						}
					}
					else {
						$prob = 0.5;
					}
					$cond{final} = 1;
					$cond{result} = $result;
					$cond{prob} = $prob;
					# en cas negatiu saltaria al seguent, tot i que en els ultims valors
					# possibles ja serà impossible
					$cond{seguent_ko} = $id+1;
				}
				elsif ($result =~ /\[(\w+)\]/) {
					$cond{final} = 0;
					$cond{seguent_ok} = $1;
					# en cas negatiu saltaria al seguent, tot i que en els ultims valors
					# possibles ja serà impossible
					$cond{seguent_ko} = $id+1;
				}
				else {
					$cond{final} = 0;
					$cond{seguent_ok} = $id+1;
					push(@pendents, $id); # falta saber on salta en cas negatiu
					$cond{seguent_ko} = -1;
				}
				
				
				
				$condicions[$id] = \%cond;
				$tincarbre = 1;
				$id++;
			}
			else {
				$dins = 0;
			}
		}
		elsif ($l =~ /^Simplified Decision Tree:/) {
			$dins = 1;
			$acumnivell = 0;
			$id = 0;
			$l = <F>; # llegim la linia en blanc
		}
		elsif ($tincarbre && $l =~ /^Subtree \[(\w+)\]/) {
# 			print "SUBTREE: $1\n" if ($VERB);
			my $st = $1;
			foreach my $c (@condicions) {
				if ($c->{seguent_ok} eq $st) {
					$c->{seguent_ok} = $id;
					$acumnivell = $c->{nivell} + 1;
					last;
				}
			}
			$dins = 1;
			$l = <F>; # llegim la linia en blanc
		}
	}
	close (F);
	
	open(RUL, ">$fArbre.perl") || die "No puc escriure $fArbre.perl";
	for (my $c = 0; $c < @condicions; $c++) {
		my %cond = %{$condicions[$c]};
		if ($cond{final}) {
			$cond{seguent_ok} = -1;
		}
		print RUL "$c|$cond{condicio}|$cond{seguent_ok}|$cond{seguent_ko}|$cond{result}|$cond{prob}\n";
	}
	close (RUL);
}

sub c45rules
{
	my ($dt, $opt) = @_;
	my $file = $dt->modelFile;
	system("$PATH_C45RULES $opt->{c45rules} -f$file > $file.rulesoutput");
	return $dt->ReadDTC45Rules("$file.rulesoutput");
}

# Extreu regles de la sortida del c4.5rules
sub ReadDTC45Rules
{
	my ($obj, $file) = @_;
	my $dins = 0;
	my @cond;
	my @regles;
	
	open (REG, $file) || die "Can not open $file";
	while (my $l = <REG>) {
		chomp($l);
		if ($dins) {
			$l =~ s/^\s*//g;
			if ($l !~ /^\-\> /) {
				my ($atribut, $condicio, $valor) = split(/ /, $l);
				
				if ($condicio eq '=') {
					if ($valor =~ /^[\.\d]+$/) {
						$condicio = '==';
					}
					else {
						$condicio = 'eq';
					}
				}
				
				if ($valor !~ /^[\.\d]+$/) {
					$valor = qq|'$valor'|;
				}
				
				my $c = '$a{' . $atribut . '}' . " $condicio " . $valor;
				push(@cond, $c);
			}
			else {
				my $regla = '$r = ' . join(' && ', @cond);
				my $val;
				if ($l =~ /class [\+1]/) {
					$val = 1;
				}
				else {
					$val = -1;
				}
				push(@regles, [$regla, $val]);
# 				print "Regla: $regla val: $val\n";
				$dins = 0;
			}
		}
		elsif ($l =~ /^Rule /) {
			$dins = 1;
			@cond = ();
		}
	}
	close (REG);
	return (@regles);
}

1;
