package Document;
our @ISA = qw(Exporter);
use strict;
use Markable;
use Mention;
use Features;

# Per defecte Document.pm s'espera un document amb COLUMNES

###############################################################################
# globals

our %FIXCOL;
our %FORMAT;

require('Document.inc');

our @EXPORT_OK = qw(%FIXCOL %FORMAT);

###############################################################################
# functions


sub new {
	my ($pkg, $file) = @_;
	my $doc = bless { _file => $file }, $pkg;
	$doc->OpenDocument;
	return $doc;
}

sub file {
	my ($doc) = @_;
	return $doc->{_file};
}

sub trueMentions {
	my ($doc, $value) = @_;
	if (defined($value)) {
		$doc->{_trueMentions} = $value;
	}
	else {
		return $doc->{_trueMentions};
	}
}

sub GetCols
{
	my $doc = shift;
	my @colNums = @_;
	my $file = $doc->file;
	
	# primer columna, despres fila
	my @matrix;
	
	open (D, "$file") || die "No puc obrir $file";
	while (my $l = <D>) {
		chomp($l);
		if ($l ne '') {
			my @cols = split(/\t/, $l);
			for (my $i = 0; $i < @colNums; $i++) {
				push(@{$matrix[$i]}, $cols[$colNums[$i]]);
			}
		}
	}
	close (D);
	
	return @matrix;
}


# Obre el document i carrega els tokens
sub OpenDocument {
	my $doc = shift;
	
	# obtenim tots els words (tokens)
	my @tokens;
	my @col = $doc->GetCols($FIXCOL{token});
	my $cc = 0; # Primer token te ID 0 !!
	foreach my $l (@{$col[0]}) {
		my $t = new Markable("word_$cc");
		$t->document($doc);
		$t->index($cc);
		$t->content($l);
		$tokens[$cc] = $t;
		$cc++;
	}
	$doc->{_tokens} = \@tokens;
}

sub tokens {
	my $doc = shift;
	return $doc->{_tokens};
}

sub GetLevelSentence
{
	my ($doc) = @_;
	my $file = $doc->file;
	my @sentences;
	my $tok = 0;
	my $initok = 0;
	my $cc = 0;
	open (D, "$file") || die "No puc obrir $file";
	while (my $l = <D>) {
		chomp($l);
		if ($l eq '') {
			if ($tok >= $initok) {
				my $m = new Markable('sentence_' . $cc);
				$m->document($doc);
				$m->index($cc);
				$m->span('word_' . $initok . '..word_' . ($tok-1));
				push(@sentences, $m);
				$initok = $tok;
				$cc++;
			}
		}
		else {
			$tok++;
		}
	}
	
	if ($tok >= $initok) {
		my $m = new Markable('sentence_' . $cc);
		$m->document($doc);
		$m->index($cc);
		$m->span('word_' . $initok . '..word_' . ($tok-1));
		push(@sentences, $m);
	}
	
	close (D);
	
	return @sentences;
}

sub GetLevelNE
{
	my ($doc, $ncol) = @_;
	my $file = $doc->file;
	my @nes;
	my $tok = 0;
	my $initok = 0;
	my @cols = $doc->GetCols($FIXCOL{enamex});
	my $cc = 0;
	if ($FORMAT{enamex} eq 'BIO') {
		my $pending = undef;
		foreach my $l (@{$cols[0]}) {
			if (defined($pending) && $l =~ /^(B-|O)/) {
				if ($initok == $tok-1) {
					$pending->span('word_' . $initok);
				}
				else {
					$pending->span('word_' . $initok . '..word_' . ($tok - 1));
				}
				push(@nes, $pending);
				$pending = undef;
			}
			
			if ($l =~ /^B-(.*?)$/) {
				my $class = $1;
				my $m = new Markable('enamex_' . $cc);
				$m->document($doc);
				$m->index($cc);
				$m->{tag} = $class;
				$initok = $tok;
				$cc++;
				$pending = $m;
			}
			$tok++;
		}
	
		if (defined($pending)) {
			if ($initok == $tok-1) {
				$pending->span('word_' . $initok);
			}
			else {
				$pending->span('word_' . $initok . '..word_' . ($tok - 1));
			}
			$pending->span('word_' . $initok . '..word_' . ($tok - 1));
			push(@nes, $pending);
		}
	}
	else {
		my @pending;
		foreach my $cInfo (@{$cols[0]}) {
			if ($cInfo ne '_') {
			
				# one-token NE
				while ($cInfo =~ s/\((\w+)\*?\)//) {
					my $class = $1;
					my $m = new Markable('enamex_' . $cc);
					$m->document($doc);
					$m->index($cc);
					$m->span('word_' . $tok);
					$m->{tag} = $class;
					$cc++;
					push(@nes, $m);
				}
				
				# begin of NE
				while ($cInfo =~ s/\((\w+)\*?//) {
					my $class = $1;
					my $m = new Markable('enamex_' . $cc);
					$m->document($doc);
					$m->index($cc);
					$m->{tag} = $class;
					$m->{initok} = $tok;
					$cc++;
					push(@pending, $m);
				}
				
				# end of NE
				while ($cInfo =~ s/\*?\)//) {
					my $ne = pop(@pending);
					if (defined($ne)) {
						$ne->span('word_' . $ne->{initok} . '..word_' . $tok);
						$ne->{initok} = undef;
						push(@nes, $ne);
					}
					else {
						die "Detected the end of a named entity without begin (?,$tok)";
					}
				}
			}
			$tok++;
		}
	}
	
	# Les ordenem
	my %last;
	@nes = sort OrderMentions (@nes);
	for (my $i = 0; $i < @nes; $i++) {
		$nes[$i]->index($i);
		$nes[$i]->id('enamex_' . $i);
	}
	
	return @nes;
}


sub GetIndex
{
	my ($ind, $i) = @_;
	if (!defined($ind->{$i})) {
		my $n = $ind->{nexti} || 0;
		$ind->{$i} = $n;
		$n++;
		$ind->{nexti} = $n;
	}
	
	return $ind->{$i};
}


sub GetLevelCoref
{
	my ($doc, $level, $column) = @_;
	my %ind;
	my $file = $doc->file;
	
	open (F, $file) || die "Can not open $file: $!";
	
	# Extract the keys from the file until #end is found
	my $lnumber = 0;
	my @mentions;
	my @half;
	my @last;
	my $id = 0;
	while (my $l = <F>) {
		chomp($l);
		next if ($l eq '');
		next if ($l =~ /^\#/);
		my @columns = split(/\t/, $l);
		my $cInfo = $columns[$column];
		if ($cInfo ne '_') {
			
			# one-token mention(s)
			while ($cInfo =~ s/\((\d+)\)//) {
				my $ie = GetIndex(\%ind, $1);
				my $mention = new Markable($level . '_' . $id);
				$mention->document($doc);
				$mention->index($id);
				$mention->{chain} = $ie;
				$mention->span("word_$lnumber");
				$last[$ie] = $mention->{id};
				$mentions[$id] = $mention;
				$id++;
			}
			
			# begin of mention(s)
			while ($cInfo =~ s/\((\d+)//) {
				my $ie = GetIndex(\%ind, $1);
				push(@{$half[$ie]}, $lnumber);
			}
			
			# end of mention(s)
			while ($cInfo =~ s/(\d+)\)//) {
				my $numberie = $1;
				my $ie = GetIndex(\%ind, $numberie);
				my $start = pop(@{$half[$ie]});
				if (defined($start)) {
					my $mention = new Markable($level . '_' . $id);
					$mention->document($doc);
					$mention->index($id);
					$mention->{chain} = $ie;
					$mention->span("word_$start..word_$lnumber");
					$last[$ie] = $mention->{id};
					$mentions[$id] = $mention;
					$id++;
				}
				else {
					die "Detected the end of a mention [$numberie]($ie) without begin (?,$lnumber)";
				}
			}
			
		}
		$lnumber++;
	}
	
	# Les ordenem
	my %last;
	@mentions = sort OrderMentions (@mentions);
	for (my $i = 0; $i < @mentions; $i++) {
		$mentions[$i]->index($i);
		$mentions[$i]->id($level . '_' . $i);
		if (defined($last{$mentions[$i]->{chain}})) {
			$mentions[$i]->{dir_antecedent} = $last{$mentions[$i]->{chain}};
		}
		$last{$mentions[$i]->{chain}} = $level . '_' . $i;
	}
	
	return @mentions;
}

sub GetLevelSRL
{
	my ($doc, $column) = @_;
	my $file = $doc->file;
	my @markables;
	
	open (D, "$file") || die "No puc obrir $file";
	my $tok = 0;
	my $id = 0;
	while (my $l = <D>) {
		chomp($l);
		if ($l ne '') {
			my @cols = split(/\t/, $l);
			my $npreds = (scalar(@cols) - $column);
			my $m = new Markable('srl_' . $id);
			$m->document($doc);
			$m->index($id);
			$m->{npreds} = $npreds;
			$m->span("word_$tok");
			
			my $offset = 0;
			for (my $i = 0; $i < $npreds; $i++) {
				my $apred = $cols[$column+$offset+$i];
				$apred =~ s/[\(\)]//og;
				if ($apred ne '*' && $apred ne '_') {
					$apred = lc($apred);
					$m->{'apred' . $i} = $apred;
				}
			}
			push(@markables, $m);
			$id++;
			$tok++;
		}
	}
	close (D);
	
	return @markables;
}

# Ho guardo com a capes, cada capa representa un sintagma
sub GetLevelSyntax
{
	my ($doc, $column) = @_;
	my $file = $doc->file;
	my @markables;
	
	open (F, $file) || die "Can not open $file: $!";
		
	my @pending;
	my $cc = 0;
	my $tok = 0;
	while (my $l = <F>) {
		chomp($l);
		next if ($l eq '');

		my @columns = split(/\t/, $l);
		my $cInfo = $columns[$column];
		if ($cInfo ne '*') {
			
			# one-token
			while ($cInfo =~ s/\((\w+)\*?\)//) {
				my $class = $1;
				my $m = new Markable('syn_' . $cc);
				$m->document($doc);
				$m->index($cc);
				$m->span('word_' . $tok);
				$m->{tag} = $class;
				$markables[$m->index] = $m;
				$cc++;
			}
			
			# begin
			while ($cInfo =~ s/\((\w+)\*?//) {
				my $class = $1;
				my $m = new Markable('syn_' . $cc);
				$m->document($doc);
				$m->index($cc);
				$m->{tag} = $class;
				$m->{initok} = $tok;
				push(@pending, $m);
				$cc++;
			}
				
			# end
			while ($cInfo =~ s/\*?\)//) {
				my $m = pop(@pending);
				if (defined($m)) {
					$m->span('word_' . $m->{initok} . '..word_' . $tok);
					$m->{initok} = undef;
					$markables[$m->index] = $m;
				}
				else {
					die "Detected the end of a syntactic component without begin (?,$tok)";
				}
			}
		}
		$tok++;
	}
	close (F);
	

	return @markables;
}

sub GetLevelPoSextra
{
	my ($doc, $col1, $col2) = @_;
	my @markables;
	
	my @cols = $doc->GetCols($col1, $col2);
	my $cc = 0;
	for (my $i = 0; $i < @{$cols[0]}; $i++) {
		my $t = new Markable('pos_' . $cc);
		$t->document($doc);
		$t->index($cc);
		$t->content($cols[0][$i]);
		$t->{tag} = $cols[0][$i];
		$t->span('word_' . $cc);
		
		# extra
		my @tags = split(/\|/, $cols[1][$i]);
		foreach my $tag (@tags) {
			my ($label, $value) = split(/\=/, $tag);
			$t->{$label} = $value;
		}
		
		$markables[$cc] = $t;
		$cc++;
	}
	
	return @markables;
}

sub TestLevel
{
	my ($doc, $level) = @_;
	
	if (defined($FIXCOL{$level}) || $level eq 'sentence') {
		return 1;
	}
	return 0;
}

sub GetLevel
{
	my ($doc, $level) = @_;
	
	if ($level ne 'sentence' && !defined($FIXCOL{$level})) {
		die "level $level ?";
	}
	
	if (!defined($doc->{_level}{$level})) {
		my @markables;
		if ($level eq 'sentence') {
			@markables = $doc->GetLevelSentence;
		}
		elsif ($level eq 'enamex') {
			@markables = $doc->GetLevelNE($FIXCOL{enamex});
		}
		elsif ($level =~ /(mention|coref|response)/) {
			@markables = $doc->GetLevelCoref($level, $FIXCOL{$level});
		}
		elsif ($level eq 'srl') {
			@markables = $doc->GetLevelSRL($FIXCOL{$level});
		}
		elsif ($level eq 'syn') {
			@markables = $doc->GetLevelSyntax($FIXCOL{$level});
		}
		elsif ($level eq 'pos' && $FORMAT{'pos'} eq 'extra') {
			@markables = $doc->GetLevelPoSextra($FIXCOL{$level}, $FIXCOL{posextra});
		}
		else {
			# span = 1 token
			my @cols = $doc->GetCols($FIXCOL{$level});
			my $cc = 0;
			foreach my $l (@{$cols[0]}) {
				my $t = new Markable($level . '_' . $cc);
				$t->document($doc);
				$t->index($cc);
				$t->content($l);
				$t->{tag} = $l;
				$t->span('word_' . $cc);
				$markables[$cc] = $t;
				$cc++;
			}
		}

		# Marquem els tokens amb la referencia corresponent (indexacio)
		my $n = $level . "_level";
		foreach my $markable (@markables) {
			foreach my $ref (@{$markable->{span}}) {
				$ref->{$n} = $markable->index;
			}
			
			# Marquem tambe el level
			$markable->{mmax_level} = $level;
		}

		$doc->{_level}{$level} = \@markables;
	}
	return $doc->{_level}{$level};
}

sub GetMentions {
	my ($doc, $trueMentions, $train) = @_;
	
	if (!defined($doc->{_mentions})) {
		my @mentions;
		my $markables;
		$doc->trueMentions($trueMentions);
		if ($trueMentions) {
			$markables = $doc->GetLevel('coref');
		}
		else {
			$markables = $doc->GetLevel('mention');
		}
		
		# We receive generic Markables, but we want Mentions
		foreach my $m (@$markables) {
			my $mention = new Mention($m->id);
			$mention->CopyFrom($m);
			$mention->trueMention($trueMentions);
			push(@mentions, $mention);
		}
		$doc->{_mentions} = \@mentions;
	}
	
	# Mentions for training have coreferential information
	if ($train) {
		$doc->MarkCorefChains;
	}
	
	return $doc->{_mentions};
}

sub MarkCorefChains
{
	my $doc = shift;
	$doc->IniCorefChains;
	my $c = $doc->{_coref};
	foreach my $m (@{$doc->{_mentions}}) {
		$c->{chain}{$m->index} = $m->{chain};
		push(@{$c->{vchains}[$m->{chain}]}, $m->index);
		if ($m->{chain} > $c->{cc}) {
			$c->{cc} = $m->{chain};
		}
	}
}


sub GetMentionById {
	my ($doc, $idMention) = @_;
	
	foreach my $m (@{$doc->{_mentions}}) {
		if ($m->id eq $idMention) {
			return $m;
		}
	}
	return undef;
}

sub GetTokenById {
	my ($doc, $idToken) = @_;
	
	foreach my $t (@{$doc->{_tokens}}) {
		if ($t->id eq $idToken) {
			return $t;
		}
	}
	return undef;
}

sub Corefer {
	my ($doc, $mencio_i, $mencio_j) = @_;
	my $c = $doc->{_coref};
	
	my $i = $mencio_i->index;
	my $j = $mencio_j->index;
	
	return (defined($c->{chain}{$i}) && defined($c->{chain}{$j}) &&
	$c->{chain}{$i} == $c->{chain}{$j}) ? 1 : 0;
}

sub IniCorefChains {
	my $doc = shift;
	my %c;
	%{$c{chain}} = ();
	@{$c{vchains}} = ();
	$c{cc} = 0;
	$doc->{_coref} = \%c;
}

sub SetSingle {
	my ($doc, $mencio) = @_;
	my $c = $doc->{_coref};
	my $i = $mencio->index;
	if (!defined($c->{chain}{$i})) {
		$c->{chain}{$i} = $c->{cc};
		$c->{cc}++;
		@{$c->{vchains}[$c->{chain}{$i}]} = ($mencio);
	}
}

sub SetCoref {
	my ($doc, $mencio_i, $mencio_j) = @_;
	my $c = $doc->{_coref};
	my $i = $mencio_i->index;
	my $j = $mencio_j->index;
	
	if (!defined($c->{chain}{$j}) && !defined($c->{chain}{$i})) {
		$c->{chain}{$j} = $c->{cc};
		$c->{cc}++;
		push(@{$c->{vchains}[$c->{chain}{$j}]}, $mencio_j);
		push(@{$c->{vchains}[$c->{chain}{$j}]}, $mencio_i);
		$c->{chain}{$i} = $c->{chain}{$j};
	}
	elsif (!defined($c->{chain}{$i})) {
		push(@{$c->{vchains}[$c->{chain}{$j}]}, $mencio_i);
		$c->{chain}{$i} = $c->{chain}{$j};
	}
	elsif (!defined($c->{chain}{$j})) {
		push(@{$c->{vchains}[$c->{chain}{$i}]}, $mencio_j);
		$c->{chain}{$j} = $c->{chain}{$i};
	}
	elsif ($c->{chain}{$j} != $c->{chain}{$i}) {
		my ($newchain, $oldchain) = ($c->{chain}{$i} < $c->{chain}{$j}) ?
		($c->{chain}{$i}, $c->{chain}{$j}) : ($c->{chain}{$j}, $c->{chain}{$i});
		
		foreach my $m (@{$c->{vchains}[$oldchain]}) {
			$c->{chain}{$m->index} = $newchain;
		}
		push(@{$c->{vchains}[$newchain]}, @{$c->{vchains}[$oldchain]});
		@{$c->{vchains}[$oldchain]} = ();
	}
}

sub GetCorefChains
{
	my $doc = shift;
	if (!defined($doc->{_coref})) {
		$doc->MarkCorefChains;
	}
	return $doc->{_coref};
}

sub OrderMentions
{
	if ($a->{span_ini} < $b->{span_ini}) {
		return -1;
	}
	elsif ($a->{span_ini} > $b->{span_ini}) {
		return 1;
	}
	else {
		return (scalar(@{$b->{span}}) <=> scalar(@{$a->{span}}));
	}
}


sub AddInfoCol
{
	my ($corcolumn, $l, $info) = @_;
	
	if (!defined($corcolumn->{$l})) {
		$corcolumn->{$l} = $info;
	}
	else {
		if ($corcolumn->{$l} =~ /^\(\d+/) {
			$corcolumn->{$l} .= '|' . $info;
		}
		else {
			$corcolumn->{$l} = $info . '|' . $corcolumn->{$l};
		}
	}
}

sub MakeColumnsResponse
{
	my ($doc, $response) = @_;
	my $file = $doc->file;
	my %corcolumn;
	
	@{$doc->{_response}} = sort OrderMentions (@$response);
	foreach my $r (@{$doc->{_response}}) {
		my $entity = $r->{chain};
		my $ini = $r->{span_ini};
		my $fi = $r->{span_fi};
		if ($ini == $fi) {
			AddInfoCol(\%corcolumn, $ini, '(' . $entity . ')');
		}
		else {
			AddInfoCol(\%corcolumn, $ini, '(' . $entity);
			AddInfoCol(\%corcolumn, $fi, $entity . ')');
		}
	}
	
	my $outFile = $file . '.out';
	open (OUT, ">$outFile") || die "No puc escriure $outFile: $!";
	open (IN, $file) || die "No puc escriure $file: $!";
	my $t = 0;
	while (my $l=<IN>) {
		chomp($l);
		if ($l ne '') {
			my @cols = split(/\t/, $l);
			if (defined($corcolumn{$t})) {
				$cols[$FIXCOL{response}] = $corcolumn{$t};
			}
			else {
				$cols[$FIXCOL{response}] = '_';
			}
			print OUT join("\t", @cols) . "\n";
			$t++;
		}
		else {
			print OUT "\n";
		}
	}
	close (IN);
	close (OUT);
}

sub SaveLevelCoref
{
	my ($doc, $mentions, $column) = @_;
	my $file = $doc->file;
	my %corcolumn;
	
	foreach my $r (sort OrderMentions (@$mentions)) {
		my $entity = $r->{chain};
		my $ini = $r->{span_ini};
		my $fi = $r->{span_fi};
		if ($ini == $fi) {
			AddInfoCol(\%corcolumn, $ini, '(' . $entity . ')');
		}
		else {
			AddInfoCol(\%corcolumn, $ini, '(' . $entity);
			AddInfoCol(\%corcolumn, $fi, $entity . ')');
		}
	}
	
	system("cp $file $file.last");
	
	open (OUT, ">$file") || die "No puc escriure $file: $!";
	open (IN, "$file.last") || die "No puc escriure $file.last: $!";
	my $t = 0;
	while (my $l=<IN>) {
		chomp($l);
		if ($l ne '') {
			my @cols = split(/\t/, $l);
			if (defined($corcolumn{$t})) {
				$cols[$column] = $corcolumn{$t};
			}
			else {
				$cols[$column] = '_';
			}
			print OUT join("\t", @cols) . "\n";
			$t++;
		}
		else {
			print OUT "\n";
		}
	}
	close (IN);
	close (OUT);
}

sub SaveLevel
{
	my ($doc, $array, $level) = @_;
	
	if ($level eq 'response') {
		$doc->MakeColumnsResponse($array);
	}
	elsif ($level eq 'mention') {
		$doc->SaveLevelCoref($array, $FIXCOL{mention});
	}
	else {
		die "funcio Document::SaveLevel obsoleta per level $level";
	}
}

sub Clear {
	my $doc = shift;
	if (defined($doc->{_coref})) {
		foreach my $chain (@{$doc->{_coref}{vchains}}) {
			undef $chain;
		}
		undef $doc->{_coref};
	}
	foreach my $t (@{$doc->{_tokens}}) {
		undef $t;
	}
	if (defined($doc->{_level})) {
		foreach my $l (%{$doc->{_level}}) {
			foreach my $m (@{$doc->{_level}{$l}}) {
				undef $m;
			}
		}
	}
	
	if (defined($doc->{_mentions})) {
		foreach my $m (@{$doc->{_mentions}}) {
			undef $m;
		}
		undef $doc->{_mentions};
	}
	if (defined($doc->{_response})) {
		foreach my $m (@{$doc->{_response}}) {
			undef $m;
		}
		undef $doc->{_response};
	}
}

1;
