package Graph;
require Model;
our @ISA = qw(Model);
use strict;
use Vertex;
use Edge;

##############################################################################
# globals


##############################################################################
# functions

sub Nprune
{
	my $obj = shift;
	if (@_) {
		$obj->{_Nprune} = shift;
	}
	else {
		return $obj->{_Nprune};
	}
}

sub GetConstraints
{
	my ($obj, $file, $noWarning) = @_;
	my %apply;
	my $fName = "$file.constraints";
	$fName .= '-' . $obj->alias if (defined($obj->alias));
	if(open (G, $fName)) { # || die "Can not open $fName";
		while (my $l = <G>) {
			chomp($l);
			my @camps = split(/:/, $l);
			my ($edge, $lc) = @camps;
			@{$apply{$edge}} = split(/,/, $lc);
		}
		close (G);
	}
	else {
		print "Warning: no constraints applied to this file ($fName)\n" unless ($noWarning);
	}
	return \%apply;
}


# Generates vertices and edges converting a document to a graph
# fd can be a file name or a Document object
sub GenGraph
{
	my ($obj, $fd, $opt) = @_;
	my $doc = (ref($fd) eq 'Document') ? $fd : new Document($fd);
	my $N = $obj->Nprune;
	my $mentions = $doc->GetMentions($obj->trueMentions);
	
	# Create vertices
	my @vertices;
	my $gi = 0;
	my @map;
	foreach my $t ('e', 's', 'p') {
		foreach my $mention (@$mentions) {
			next if ($mention->type ne $t);
			my $vertex = new Vertex($mention);
			$vertex->graphid($gi);
			$map[$mention->index] = $gi;
			$gi++;
			push(@vertices, $vertex);
		}
	}
	$obj->{mapMentions} = \@map;
	
	my $apply = $obj->GetConstraints($doc->file);
	
	# Create edges (each vertex is connected with all the others)
	my @edges;
	for (my $j = 1; $j < @vertices; $j++) {
		my @adjacents;
		my ($hihaPos, $hihaNeg) = (0, 0);
		for (my $i = $j-1; $i >= 0; $i--) {
			my $edge = new Edge($vertices[$i], $vertices[$j]);
			$edge->{_model} = $obj;
			
			my ($mi, $mj) = ($vertices[$i]->id < $vertices[$j]->id) ?
			($vertices[$i]->id, $vertices[$j]->id) : ($vertices[$j]->id, $vertices[$i]->id);
			my $edge_str = $mi . ',' . $mj;
			my $weight = 0;
			foreach my $c (@{$apply->{$edge_str}}) {
				$weight += $obj->weights->[$c];
			}
			
			$hihaPos++ if ($weight > 0);
			$hihaNeg++ if ($weight < 0);
			$edge->weight($weight);
			push(@adjacents, $edge);
		}
		
		# Pruning, let the N most significative edges and discard the others
		if ($N && @adjacents > $N) {
			@adjacents = sort { $a->weight <=> $b->weight } @adjacents;
			my $selPos = ($hihaPos > $N/2) ? $N/2 : $hihaPos;
			my $selNeg = ($hihaNeg > $N - $selPos) ? $N - $selPos : $hihaNeg;
			$selPos = $N - $selNeg if ($selPos + $selNeg < $N);
			if ($selPos) {
				splice(@adjacents, $selNeg, -$selPos);
			}
			else {
				splice(@adjacents, $selNeg);
			}
		}
		
		push(@edges, @adjacents);
	}
	
	# Each vertex has a refrence to the edges
	foreach my $v (@vertices) {
		$v->edges(\@edges);
	}
	
	print scalar(@vertices) . " vertices\n";
	
	# In case that there are extra models (entity-mention ones) this is the
	# time to incorporate them to the graph. Graph becomes hypergraph
	if (defined($opt->{extraModels}) && @{$opt->{extraModels}}) {
		foreach my $eModel (@{$opt->{extraModels}}) {
			my $cs = $eModel->constraints;
			my $csWeights = $eModel->weights;
			my $eApply = $eModel->GetConstraints($doc->file, 1);
			foreach my $hyperedge (keys(%{$eApply})) {
				my @am = split(/\,/, $hyperedge);
				foreach my $cc (@{$eApply->{$hyperedge}}) {
					next if ($csWeights->[$cc] == 0); # pesos a 0 vol dir que l'hyperedge no aplica
					my ($cond, $constraint) = @{$cs->[$cc]};
					my $getsInfluenced = $am[$constraint->{reb}];
# 					print "getsInfluenced: m$getsInfluenced v$map[$getsInfluenced]\n";
					my %myc;
					
					# adjacencies
					for (my $ii = 0; $ii < @am; $ii++) {
						if ($ii != $constraint->{reb}) {
							push(@{$myc{adj}}, $map[$am[$ii]]);
# 							print "Adjacent: $ii m$am[$ii] v$map[$am[$ii]]\n";
						}
					}
					
					# influence rule
					for (my $ii = 0; $ii < @{$constraint->{inf}}; $ii++) {
						for (my $jj = 0; $jj < @{$constraint->{inf}[$ii]}; $jj++) {
							$myc{inf}[$ii][$jj] = $map[$am[$constraint->{inf}[$ii][$jj]]];
# 							print "Influence $ii $jj: $constraint->{inf}[$ii][$jj] m$am[$constraint->{inf}[$ii][$jj]] v$map[$am[$constraint->{inf}[$ii][$jj]]]\n";
						}
					}
					$myc{weight} = $csWeights->[$cc];
					
					my $c2 = $vertices[$map[$getsInfluenced]]->c2;
					push(@$c2, \%myc);
					$vertices[$map[$getsInfluenced]]->c2($c2);
					
					# Recover the pruned edges that relates these vertices
# 					foreach my $vAdj (@{$myc{adj}}) {
# 						if (!$vertices[$map[$getsInfluenced]]->isAdjacent($vertices[$vAdj])) {
# 							my $edge = new Edge($vertices[$map[$getsInfluenced]], $vertices[$vAdj]);
# 							$edge->{_model} = $obj;
# 				
# 							my ($mi, $mj) = ($getsInfluenced < $vertices[$vAdj]->id) ?
# 							($getsInfluenced, $vertices[$vAdj]->id) : ($vertices[$vAdj]->id, $getsInfluenced);
# 							my $edge_str = $mi . ',' . $mj;
# 							my $weight = 0;
# 							foreach my $c (@{$apply->{$edge_str}}) {
# 								$weight += $obj->weights->[$c];
# 							}
# 				
# 							$edge->weight($weight);
# 							push(@edges, $edge);
# 							$vertices[$map[$getsInfluenced]]->{_adjacencies} = undef;
# 						}
# 					}
				}
			}
		}
	}
	
	# free memory
	$doc->Clear unless (ref($fd) eq 'Document');
	
	$obj->{_vertices} = \@vertices;
	$obj->{_edges} = \@edges;
	
	return (\@vertices, \@edges);
}

sub vertices
{
	my $obj = shift;
	return $obj->{_vertices}
}

sub edges
{
	my $obj = shift;
	return $obj->{_edges}
}


# Avoid cross-references
sub Clear {
	my $obj = shift;
	
	# delete references to vertices
	foreach my $e (@{$obj->{_edges}}) {
		undef($e->{_a});
		undef($e->{_b});
	}
	undef($obj->{_edges});
	
	# delete references to edges
	foreach my $v (@{$obj->{_vertices}}) {
		undef($v->{_edges});
		undef($v->{_adjacencies});
		undef($v->{_c2});
	}
	undef($obj->{_vertices});
}

1;
