package MentionDetection;

use Document;
use Features;
use strict;

sub OrderMentions
{
	if ($a->{span_ini} < $b->{span_ini}) {
		return -1;
	}
	elsif ($a->{span_ini} > $b->{span_ini}) {
		return 1;
	}
	else {
		return (scalar(@{$b->{span}}) <=> scalar(@{$a->{span}}));
	}
}


sub SubDepTree
{
	my ($index, $dep, $pos, $dir) = @_;
	my $result = undef;
	my $doc = $dep->[0]->document;
	my $sentences = $doc->GetLevel('sentence');
	my $sindex = $dep->[$index]->{span}[0]->{sentence_level};
	my ($inici, $fi) = ($sentences->[$sindex]->{span_ini}, $sentences->[$sindex]->{span_fi});
	my $offset = $inici - 1;
	
	if ($dir == -1) {
		# search to the left
		$result = $index;
		for (my $i = $index - 1; $i >= $inici; $i--) {
			if ($dep->[$i]->{tag} + $offset == $index && $pos->[$i]->{tag} !~ /[\.\,]/) {
				$result = $i;
			}
# 			elsif ($pos->[$i]->{tag} =~ /[\.\,]/) {
# 				last;
# 			}
		}
	}
	else {
		# search to the right
		$result = $index;
		for (my $i = $index + 1; $i <= $fi ; $i++) {
			if ($dep->[$i]->{tag} + $offset == $index && $pos->[$i]->{tag} !~ /[\.\,]/) {
				$result = $i;
			}
# 			elsif ($pos->[$i]->{tag} =~ /[\.\,]/) {
# 				last;
# 			}
		}
	}
	
	# Cridem la funcio reciprocament
	my $rant = $index;
	while ($rant != $result) {
		$rant = $result;
		$result = SubDepTree($rant, $dep, $pos, $dir);
	}
	
	return $result;
}

sub DepTree
{
	my ($index, $dep, $pos) = @_;
	
	my $ini = SubDepTree($index, $dep, $pos, -1); # -1 esquerra
	my $fi = SubDepTree($index, $dep, $pos, 1); # 1 dreta
	
	return ($ini, $fi);
}

sub DepChunks
{
	my $doc = shift;
	my $dep = $doc->GetLevel('dep');
	my $pos = $doc->GetLevel('pos');
	
	my @chunks;
	my $lastNoun = undef;
	my $mention = 0;
	for (my $i = 0; $i < @$dep; $i++) {
		if ($pos->[$i]->{tag} =~ /^NN/ || $pos->[$i]->{tag} eq 'n') {
			$lastNoun = $i;
		}
		elsif (defined($lastNoun)) {
			my ($iniM, $fiM) = DepTree($lastNoun, $dep, $pos);
			my $span = ($iniM == $fiM) ? "word_$iniM" : "word_$iniM..word_$fiM";
			my $chunk = new Mention('mention_' . $mention);
			$chunk->document($doc);
			$chunk->span($span);
			push(@chunks, $chunk);
			$mention++;
			$lastNoun = undef;
		}
	}
	
	if (defined($lastNoun)) {
		my ($iniM, $fiM) = DepTree($lastNoun, $dep, $pos);
		my $span = ($iniM == $fiM) ? "word_$iniM" : "word_$iniM..word_$fiM";
		my $chunk = new Mention('mention_' . $mention);
		$chunk->document($doc);
		$chunk->span($span);
		push(@chunks, $chunk);
		$mention++;
		$lastNoun = undef;
	}	
	
	return @chunks;
}


# Limit mentions to NE, pronouns, and NPs that:
# - Are alias of an NE
# - the head is repeated at least once
#NO - definites and short (no comas, no other NPs embedded)
sub ReduceMentions
{
	my ($doc, $mentions, $opt) = @_;
	
	my @NEs;
	my %heads;
	my @save;
	# store the NEs, and the NPs with repeated heads
	for (my $i = 0; $i < @$mentions; $i++) {
		my $m = $mentions->[$i];
		if ($m->isEntity) {
			push(@NEs, $m);
		}
		elsif ($m->type eq 's') {
			my $h = lc( $m->head_str );
			if (defined($heads{$h})) {
				if (ref($heads{$h}) eq 'Mention') {
					$save[$heads{$h}->index] = 1;
					$heads{$h} = 1;
				}
				$save[$i] = 1;
			}
			else {
				$heads{$h} = $m;
			}
		}
	}
	
	# Compare NPs and NEs
	for (my $i = 0; $i < @$mentions; $i++) {
		my $m = $mentions->[$i];
		if (!$save[$i] && $m->type eq 's') {
			foreach my $ne (@NEs) {
				if (&Features::Alias( $m, $ne) == 1) {
					$save[$i];
				}
			}
		}
	}
	
	my @newmentions;
	for (my $i = 0; $i < @$mentions; $i++) {
		my $m = $mentions->[$i];
		if ($save[$i] || $m->type ne 's' || $m->isProperName) {
			push(@newmentions, $m);
		}
	}
	
	# Reorder
	@newmentions = sort OrderMentions (@newmentions);
	for (my $i = 0; $i < @newmentions; $i++) {
		$newmentions[$i]->id('mention_' . $i);
		$newmentions[$i]->index($i);
	}
	
	return @newmentions;
}

# Entrada: doc
# Repassar: li he afegit el PoSEXTRA
sub Detect
{
	my ($doc, $opt) = @_;
	my $tokens = $doc->tokens;
	my $pos = $doc->GetLevel('pos');
	my $enamex = $doc->GetLevel('enamex');
	my @mencions;
	my $ii = 0;
	
	# Add NPs
	if ($doc->TestLevel('syn')) {
		my $syn = $doc->GetLevel('syn');
		for (my $i = 0; $i < @$syn; $i++) {
			my $np = $syn->[$i];
			if ($np->{tag} eq 'NP' || $np->{tag} eq 'NML') {
				my $id = 'mention_' . $ii;
				my $copia = new Mention($id);
				$copia->CopyFrom($np);
				$copia->id($id);
				$copia->index($ii);
				push(@mencions, $copia);
				$ii++;
			}
		}
	}
	else {
		@mencions = DepChunks($doc);
		$ii = scalar(@mencions);
	}
	
	# Add pronouns and possessive pronouns
	foreach my $t (@$tokens) {
		if ($pos->[$t->index]->{tag} =~ /^(PRP\$?|WP\$|POS)$/ ||
			$pos->[$t->index]->{postype} eq 'possessive') {
			my $men = new Mention('mention_' . $ii);
			$men->document($doc);
			$men->span($t->id);
			$ii++;
			push(@mencions, $men);
		}
	}
	
	# Add Named Entities
	foreach my $ne (@$enamex) {
		my $men = new Mention('mention_' . $ii);
		$men->document($doc);
		$men->span($ne->{span_text});
		$ii++;
		push(@mencions, $men);
	}
	
	# Add capitalized nouns and proper names that appear more than 1 time in the document.
	my %marca;
	foreach my $t (@$tokens) {
		if ((($pos->[$t->index]->{tag} =~ /^NNP?S?$/ || $pos->[$t->index]->{tag} eq 'n') && $t->content =~ /^[A-Z]/) || $t->content =~ /^\d+$/) {
			push(@{$marca{$t->content}}, $t);
		}
	}
	foreach my $key (%marca) {
		if (defined($marca{$key}) && @{$marca{$key}} > 1) {
			foreach my $t (@{$marca{$key}}) {
				my $men = new Mention('mention_' . $ii);
				$men->document($doc);
				$men->span($t->id);
				$ii++;
				push(@mencions, $men);
			}
		}
	}
	
	# Discard repeated mentions
	my @elimina;
	my $mant = undef;
	foreach my $m (sort { $a->{span_text} cmp $b->{span_text} }(@mencions)) {
		print $m->index . " " . $m->{span_text} . "\n";
		if (defined($mant) && $m->{span_text} eq $mant->{span_text}) {
			push(@elimina, $mant->index);
		}
		$mant = $m;
	}
	foreach my $i (sort { $b <=> $a } (@elimina)) {
		splice(@mencions, $i, 1);
	}
	
	
	# Reorder
	for (my $i = 0; $i < @mencions; $i++) {
		$mencions[$i]->id('mention_' . $i);
		$mencions[$i]->index($i);
	}
	
	@elimina = ();
	$mant = undef;
	# In case that two mentions share the same head, the small one is discarded
	foreach my $m (sort { $a->head->id cmp $b->head->id }(@mencions)) {
		if (defined($mant) && $m->head->id eq $mant->head->id) {
			my ($llarg, $curt) = (scalar(@{$m->{span}}) >= scalar(@{$mant->{span}})) ?
										($m, $mant) : ($mant, $m);
			
			# Keep the largest and discart the shortest, except:
			#  - when the largest has ANDs or ORs (composite)
			#  - when the shortest is is a PNe
			#  - when the shortest is a pronoun
			#  - when the shortest is a NE and the largest is really longer
			
			my $isComposite = $m->isComposite;
			my $PNe = 0;
			if ($m->isEntity || $m->isProperName) {
				for (my $i = 0; $i < @mencions; $i++) {
					if ($mencions[$i]->index != $m->index &&
						&Features::StringMatch($mencions[$i], $m) == 1) {
						$PNe = 1;
						last;
					}
				}
			}
			my $toolong = 0;
			if ($curt->isEntity && (scalar(@{$llarg->{span}}) - scalar(@{$curt->{span}}) > 5)) {
				$toolong = 1;
				# discard the NP in this case
				push(@elimina, $llarg->index);
			}
			
			if (!$isComposite && !$PNe && !$m->isPronoun && !$toolong) {
				push(@elimina, $curt->index);
			}
			
			
		}
		$mant = $m;
	}
	
	foreach my $i (sort { $b <=> $a } (@elimina)) {
		splice(@mencions, $i, 1);
	}
	
	# Reorder
	@mencions = sort OrderMentions (@mencions);
	for (my $i = 0; $i < @mencions; $i++) {
		$mencions[$i]->id('mention_' . $i);
		$mencions[$i]->index($i);
	}
	
	if (defined($opt->{reduceMentions})) {
		@mencions = ReduceMentions($doc, \@mencions, $opt);
	}
	
	# Consider all of them as singletons
	for (my $i = 0; $i < @mencions; $i++) {
		$mencions[$i]->{chain} = $i;
	}
	
	$doc->SaveLevel(\@mencions, 'mention');
	
	return @mencions;
}

# annotates gold information for training from the true mentions column
sub AnnotateGold
{
	my ($doc, $docTrue) = @_;
	
	# Match true mentions with the ones detected
	my $match = 0;
	my $mentions = $doc->GetMentions(0);
	my $coref = $docTrue->GetMentions(1);
	my @marca;
	my $maxChain=0;
	for (my $i = 0; $i < @$mentions; $i++) {
		$mentions->[$i]->{chain} = undef;
		for (my $j = 0; $j < @$coref; $j++) {
			# Strict boundary matching
			if ($mentions->[$i]->{span_text} eq $coref->[$j]->{span_text}) {
				$mentions->[$i]->{chain} = $coref->[$j]->{chain};
				$match++;
				$marca[$j] = 1;
				if ($mentions->[$i]->{chain} > $maxChain) {
					$maxChain = $mentions->[$i]->{chain};
				}
				last;
			}
		}
	}
	
	for (my $j = 0; $j < @$coref; $j++) {
		if (!$marca[$j]) {
			print "Missed: " . $coref->[$j]->string . '(' . $coref->[$j]->{span_text} . ")\n";
		}
	}
	
	# annotate singletons with different chain numbers
	for (my $i = 0; $i < @$mentions; $i++) {
		if (!defined($mentions->[$i]->{chain})) {
			$maxChain++;
			$mentions->[$i]->{chain} = $maxChain;
		}
	}
	
	$doc->SaveLevel($mentions, 'mention');
	
	my $pre = @$mentions ? $match / scalar(@$mentions) : 0;
	my $rec = @$coref ? $match / scalar(@$coref) : 0;
	my $f1 = ($rec + $pre) ? 2*$rec*$pre/($rec+$pre) : 0;
	print "Mention detection: " . scalar(@$mentions) . " ($match de " . scalar(@$coref) . ")\n";
	print "P: $pre, R: $rec, F1: $f1\n";

	return ($match, scalar(@$mentions), scalar(@$coref));
}

1;
