#include "RelaxGraph.h" 

void PushConstraint( Node *n, Constraint c)
{
	n->constraints.push_back(c);
}

void PushConstraint2( Node *n, Constraint2 c)
{
	n->constraints2.push_back(c);
}

void PushAdjC2( Constraint2 *c, int v)
{
	c->adj.push_back(v);
}

void PushInfC2( Constraint2 *c, int i, int v)
{
	if (c->inf.size() < i+1) {
		c->inf.resize(i+1);
		c->inf[i].resize(1);
	}
	c->inf[i].push_back(v);
}

void PushIniLabel( Node *n, double l)
{
	n->iniLabels.push_back(l);
}

double ReadCurrentLabel( Problem *p, int node, int index)
{
// 	return p->nodes[node].iniLabels[index];
	return (*(p->current))[node][index];
}

void SetCurrentLabel( Problem *p, int node, int index, double value)
{
	//p->nodes[node].iniLabels[index] = value;
	(*p->current)[node][index] = value;
}

void PushNode( Problem *p, Node n)
{
	p->nodes.push_back(n);
}

int ReadNodeLabel( Problem *p, int index)
{
	return p->nodes[index].label;
}

void InitializeGraph(vector<Node> &nodes,
											 vector<vector<double> > &labels_a,
											 vector<vector<double> > &labels_b)
{
	cout << "Iniciant labels" << endl;  
	
	labels_a.resize(nodes.size());
	labels_b.resize(nodes.size());

	for (int i = 0; i < nodes.size(); i++) {
		int nLabels = nodes[i].iniLabels.size();
		labels_a[i].resize(nLabels);
		labels_b[i].resize(nLabels);
		for (int l = 0; l < nLabels; l++)
			labels_a[i][l] = nodes[i].iniLabels[l];
	}
}


double NormalizeSupport(double x)
{
	//TODO: posar ScaleFactor com a variable d'entrada
	//double ScaleFactor = 4096.0;
	//double ScaleFactor = 1024.0;
	//double ScaleFactor = 512.0;
	//double ScaleFactor = 128.0;
	
	//fins ara 16.0 anava be per les corefs
	//double ScaleFactor = 64.0;
	double ScaleFactor = 8.0;
	
	if ((x > ScaleFactor) || (x < -ScaleFactor)) {
		if (x > 0) x = ScaleFactor;
		else x = -ScaleFactor;
	}
	
	return (x/ScaleFactor);
}

int MaxLabel(const vector<double> &x)
{
	double max = x[0];
	int label=0;
	
	for (int i = 1; i < x.size(); i++) {
		if (x[i] >= max) {
			max = x[i];
			label = i;
		}
	}
	
	return label;
}  


void Relaxation(Problem *p)
{
	vector<Node> *nodes = &(p->nodes);
	double *support;
	double fnorm = 0.0;
	
	vector<vector<double> > labels_a, labels_b;
	vector<vector<double> > *currentLabels, *nextLabels, *swap;
	
	currentLabels = &labels_a;
	nextLabels = &labels_b;
	
	//Initialize the graph
	InitializeGraph(*nodes, labels_a, labels_b);
	
	//Loop. iterate until convergence (no changes)
	cout << "Solving..." << endl;
	int n = 0;
	bool canvis = true;
	int MaxIter = 4000; //TODO: Fer MaxIter com a parametre d'entrada
	//TODO: posar Epsilon com a parametre
	double Epsilon = 0.001;
	double singleFactor = 0.1;
	
	while (canvis && n < MaxIter) {
		canvis = false;
		
		// Per cada vertex
		for (int i = 0; i < nodes->size(); i++) {
			
			int nLabels = (*currentLabels)[i].size();
			support = new double[nLabels];
			for (int l = 0; l < nLabels; l++)
				support[l] = 0.0;
			
			// Per cada adjacent
			for (int j = 0; j < (*nodes)[i].constraints.size(); j++) {
				int vei_j = (*nodes)[i].constraints[j].adj;
				double pes_j = (*nodes)[i].constraints[j].weight;
				int nlVei = (*currentLabels)[vei_j].size();
				
				// Per cada label
				for (int l = 0; l < nLabels && l < nlVei; l++) {
					// Calculem support per label l
					support[l] += pes_j * (*currentLabels)[vei_j][l];
				}
			}
			
			// Tendencia cap a singleton
			support[nLabels-1] = singleFactor;
			
// 			// Restriccions de major ordre
// 			for (int j = 0; j < (*nodes)[i].constraints2.size(); j++) {
// 				double pes_j = (*nodes)[i].constraints2[j].weight;
// 				double *inf = new double[nLabels];
// 				for (int l = 0; l < nLabels; l++) inf[l] = 1.0;
// 				
// 				int minLabels = nLabels;
// 				// Per cada vertex adjacent implicat en la restriccio d'ordre > 1
// 				for (int k = 0; k < (*nodes)[i].constraints2[j].adj.size(); k++) {
// 					int vei_k = (*nodes)[i].constraints2[j].adj[k];
// 					int nlVei = (*currentLabels)[vei_k].size();
// 					
// 					if (nlVei < minLabels)
// 						minLabels = nlVei;
// 					
// 					// Per cada label
// 					for (int l = 0; l < minLabels; l++)
// 						inf[l] *= (*currentLabels)[vei_k][l];
// 				}
// 				
// 				// Per cada label
// 				for (int l = 0; l < minLabels; l++)
// 					support[l] += pes_j * inf[l];
// 				
// 				delete inf;
// 			}
			
			// Restriccions de major ordre
			for (int j = 0; j < (*nodes)[i].constraints2.size(); j++) {
				double pes_j = (*nodes)[i].constraints2[j].weight;
				if (pes_j) {
					for (int l = 0; l < nLabels; l++) {
						double inf = 1.0;
						for (int ii = 0; ii < (*nodes)[i].constraints2[j].inf.size(); ii++) {
							double tinf = 1.0;
							
							for (int jj = 0; jj < (*nodes)[i].constraints2[j].inf[ii].size(); jj++) {
								int vei_kk = (*nodes)[i].constraints2[j].inf[ii][jj];
								if ((*currentLabels)[vei_kk].size() > l) {
									tinf *= (*currentLabels)[vei_kk][l];
								} else {
									tinf *= 0.0;
								}
							}
							
							if (ii != 0) {
								tinf = 1.0 - tinf;
							}
							
							inf *= tinf;
						}
						support[l] += pes_j * inf;
					}
				}
			}
			
			// Normalitzem
			fnorm = 0.0;
			for (int l = 0; l < nLabels; l++) {
				support[l] = NormalizeSupport(support[l]);
				fnorm += (*currentLabels)[i][l] * (1 + support[l]);
			}
			
			// Actualitzem labels
			for (int l = 0; l < nLabels; l++) {
				if (fnorm) {
					(*nextLabels)[i][l] = (*currentLabels)[i][l] * (1 + support[l]) / fnorm;
				}
				else {
					(*nextLabels)[i][l] = (*currentLabels)[i][l];
				}
				
				if (!canvis) {
					double dif = (*nextLabels)[i][l] - (*currentLabels)[i][l];
					if ( dif < 0 && -dif > Epsilon)
						canvis = true;
					else if (dif > 0 && dif > Epsilon)
						canvis = true;
				}
			}
			
			delete support;
		}
		
		//valors NEXT passan a ser CURRENT
		swap = currentLabels;
		currentLabels = nextLabels;
		nextLabels = swap;
		
		n++;
		cout << ".";
		cout << flush;
	}
	cout << endl;
	
	//Assign colors. bestLabel
	for (int i = 0; i < nodes->size(); i++) {
		int label = MaxLabel((*currentLabels)[i]);
		(*nodes)[i].label = label;
	}
}

//Intent de execucio per fases
void RelaxationIni(Problem *p)
{
	vector<Node> *nodes = &(p->nodes);
	vector<vector<double> > *currentLabels, *nextLabels, *swap;
	
	currentLabels = &(p->labels_a);
	nextLabels = &(p->labels_b);
	p->current = currentLabels;
	
	//Initialize the graph
	InitializeGraph(*nodes, p->labels_a, p->labels_b);
}

int RelaxationIter(Problem *p, int MaxIter)
{
	vector<Node> *nodes = &(p->nodes);
	double *support;
	double fnorm = 0.0;
	
	vector<vector<double> > *currentLabels, *nextLabels, *swap;
	
	currentLabels = &(p->labels_a);
	nextLabels = &(p->labels_b);
	p->current = currentLabels;
	
	//Loop. iterate until convergence (no changes)
	//cout << "Solving..." << endl;
	int n = 0;
	bool canvis = true;
	//TODO: posar Epsilon com a parametre
	double Epsilon = 0.001;
	double singleFactor = 0.1;
	
	while (canvis && n < MaxIter) {
		canvis = false;
		
		// Per cada vertex
		for (int i = 0; i < nodes->size(); i++) {
			
			int nLabels = (*currentLabels)[i].size();
			support = new double[nLabels];
			
			// valor inicial
			for (int l = 0; l < nLabels; l++)
				support[l] = 0.0;
			
			// Per cada adjacent
			for (int j = 0; j < (*nodes)[i].constraints.size(); j++) {
				int vei_j = (*nodes)[i].constraints[j].adj;
				double pes_j = (*nodes)[i].constraints[j].weight;
				int nlVei = (*currentLabels)[vei_j].size();
				
				// Per cada label
				for (int l = 0; l < nLabels && l < nlVei; l++) {
					// Calculem support per label l
					support[l] += pes_j * (*currentLabels)[vei_j][l];
				}
			}
			
			// Tendencia cap a singleton
			support[nLabels-1] = singleFactor;
			
			// Restriccions de major ordre
			for (int j = 0; j < (*nodes)[i].constraints2.size(); j++) {
				double pes_j = (*nodes)[i].constraints2[j].weight;
				double *inf = new double[nLabels];
				for (int l = 0; l < nLabels; l++) inf[l] = 1.0;
				
				int minLabels = nLabels;
				// Per cada vertex adjacent implicat en la restriccio d'ordre > 1
				for (int k = 0; k < (*nodes)[i].constraints2[j].adj.size(); k++) {
					int vei_k = (*nodes)[i].constraints2[j].adj[k];
					int nlVei = (*currentLabels)[vei_k].size();
					
					if (nlVei < minLabels)
						minLabels = nlVei;
					
					// Per cada label
					for (int l = 0; l < minLabels; l++)
						inf[l] *= (*currentLabels)[vei_k][l];
				}
				
				// Per cada label
				for (int l = 0; l < minLabels; l++)
					support[l] += pes_j * inf[l];
			}
			
			// Normalitzem
			fnorm = 0.0;
			for (int l = 0; l < nLabels; l++) {
				support[l] = NormalizeSupport(support[l]);
				fnorm += (*currentLabels)[i][l] * (1 + support[l]);
			}
			
			// Actualitzem labels
			for (int l = 0; l < nLabels; l++) {
				if (fnorm) {
					(*nextLabels)[i][l] = (*currentLabels)[i][l] * (1 + support[l]) / fnorm;
				}
				else {
					(*nextLabels)[i][l] = (*currentLabels)[i][l];
				}
				
				if (!canvis) {
					double dif = (*nextLabels)[i][l] - (*currentLabels)[i][l];
					if ( dif < 0 && -dif > Epsilon)
						canvis = true;
					else if (dif > 0 && dif > Epsilon)
						canvis = true;
				}
			}
			
			delete support;
		}
		
		//valors NEXT passan a ser CURRENT
		swap = currentLabels;
		currentLabels = nextLabels;
		nextLabels = swap;
		p->current = currentLabels;
		
		n++;
		cout << ".";
		cout << flush;
	}
// 	cout << endl;
	
	//Assign colors. bestLabel
	for (int i = 0; i < nodes->size(); i++) {
		int label = MaxLabel((*currentLabels)[i]);
		(*nodes)[i].label = label;
	}
	
	if (canvis) {
		return 1;
	}
	return 0;
}
