package SolverDT;
require Solver;
our @ISA = ("Solver");
use strict;
use DT;
use Document;

use constant LINK_FIRST => 0;
use constant LINK_BEST => 1;

sub model
{
	my ($obj) = @_;
	my $model = new DT($obj->{_modelFile});
	$model->mask($obj->mask);
	$model->alias($obj->modelAlias);
	$model->trueMentions($obj->trueMentions);
	return $model;
}

sub strategy
{
	my $obj = shift;
	if (@_) {
		$obj->{_strategy} = shift;
	}
	else {
		if (!defined($obj->{_strategy})) {
			$obj->{_strategy} = LINK_FIRST;
		}
		return $obj->{_strategy};
	}
}

sub Solve
{
	my ($dt, $fitxer, $opt) = @_;
	my $doc = new Document($fitxer);
	my $mentions = $doc->GetMentions($dt->$trueMentions);
	
	print "Solving $fitxer...\n" if ($dt->verbosity);
	
	for (my $j = 0; $j < scalar(@$mentions); $j++) {
		my $mj = $mentions->[$j];
		$mj->{chain} = $j; # singleton by default
		
		my $pmax = 0;
		my $imax;
		for (my $i = $j-1; $i >= 0; $i--) {
			my $mi = $mentions->[$i];
			my ($result, $prob) = $dt->MyConsult($mi, $mj);
			
			# result 0 = '+' (corefer)
			if ($result == 0) {
				if ($dt->strategy == LINK_BEST) {
					if ($prob > $pmax) {
						$pmax = $prob;
						$imax = $i;
						last if ($prob == 1);
					}
				}
				elsif ($dt->strategy == LINK_FIRST) {
					$mj->{chain} = $mi->{chain};
					last;
				}
			}
		}
		
		# link with best one
		if ($dt->strategy == LINK_BEST && $pmax > 0) {
			$mj->{chain} = $mentions->[$imax]->{chain};
		}
	}
	
	$doc->SaveLevel($mentions, 'response');
	
	# free memory
	$doc->Clear;
}


1;
