package SolverRelax;
require Solver;
our @ISA = qw(Solver);
require Attributes;
use strict;
use Graph;
use DT;

use FindBin qw( $RealBin );
use lib "$RealBin/lib/RelaxGraph-src/blib/lib";
use lib "$RealBin/lib/RelaxGraph-src/blib/arch";
use RelaxGraph;

sub epsilon {
	my $obj = shift;
	@_ ? $obj->{_epsilon} = shift : $obj->{_epsilon};
}

sub maxIter {
	my $obj = shift;
	@_ ? $obj->{_maxIter} = shift : $obj->{_maxIter};
}

sub scaleFactor {
	my $obj = shift;
	@_ ? $obj->{_scaleFactor} = shift : $obj->{_scaleFactor};
}

sub singleFactor {
	my $obj = shift;
	@_ ? $obj->{_singleFactor} = shift : $obj->{_singleFactor};
}

# TODO: posar aquest valors com a parametres del config
sub Initialize
{
	my ($obj, $vertices, $opt) = @_;
	
	$opt->{scaleFactor} = 8 if (!defined($opt->{scaleFactor}));
	$obj->scaleFactor($opt->{scaleFactor});
	$obj->epsilon(0.001);
	$obj->maxIter(2000);
	if (defined($opt->{singleFactor})) {
		$obj->singleFactor($opt->{singleFactor});
	}
	else {
		$obj->singleFactor(0.05);
	}
	
	my $acum = 0;
	my $ini = 0;
	my $totalv = scalar(@$vertices);
	
	if (defined($opt->{nVertex})) {
		$totalv = $opt->{nVertex};
		if (defined($opt->{offset}) && $opt->{offset} > 0) {
			$ini = $opt->{offset};
			$acum = $vertices->[$ini-1]->nLabels;
		}
	}
	
	for (my $i = $ini; $i < $totalv; $i++) {
		my $vertex = $vertices->[$i];
		
		# number of labels
		$acum++; # if ($t ne 'p');
		$vertex->nLabels($acum);
		
		# equiprobable
		for (my $l = 0; $l < $vertex->nLabels; $l++) {
			# equiprobable
			$vertex->label->[$l] = 1 / ($vertex->nLabels + 1);
			$vertex->nextLabel->[$l] = 1 / ($vertex->nLabels + 1);
			
			# except the last one
			if ($l == $vertex->nLabels - 1 && $vertex->mention->type ne 'p') {
				$vertex->label->[$l] = 2 / ($vertex->nLabels + 1);
				$vertex->nextLabel->[$l] = 2 / ($vertex->nLabels + 1);
			}
		}
	}

}

sub min
{
	my ($a, $b) = @_;
	return $a if ($a <= $b);
	return $b;
}

sub Normalize
{
	my ($obj, @support) = @_;
	foreach my $s (@support) {
		$s = $obj->scaleFactor if ($s > $obj->scaleFactor);
		$s = -$obj->scaleFactor if ($s < -$obj->scaleFactor);	
		$s /= $obj->scaleFactor;
	}
	return @support;
}

sub Solve
{
	my ($obj, $doc, $opt) = @_;
	
	# Generates the graph
	my $graph = new Graph($obj->model->modelFile);
	$graph->alias($obj->model->alias);
	$graph->Nprune($opt->{Nprune}) if (defined($opt->{Nprune}));
	$graph->trueMentions($obj->model->trueMentions);
	
	my ($vertices, $edges) = $graph->GenGraph($doc, $opt);
	
	print "Solving: " . $doc->file . "...\n" if ($obj->verbosity);
	
	# Set initial state
	$obj->Initialize($vertices, $opt);
	
	# Create @nodes and send them to the RelaxGraph library
	my $tnodes = 0;
	my $problem = new RelaxGraph::Problem;
	foreach my $v (@$vertices) {
		my $n = new RelaxGraph::Node;
		foreach my $l (@{$v->label}) {
			&RelaxGraph::PushIniLabel( $n, $l);
		}
		foreach my $adj (@{$v->adjacencies}) {
			my $w = $adj->weight;
			my $neighbor = $adj->destination($v);
			next if ($neighbor->nLabels > $v->nLabels);
			my $c = new RelaxGraph::Constraint;
			&RelaxGraphc::Constraint_adj_set( $c, $neighbor->graphid);
			&RelaxGraphc::Constraint_weight_set( $c, $w);
			&RelaxGraph::PushConstraint( $n, $c);
		}
		
		# Restriccions de grup
		foreach my $c2 (@{$v->c2}) {
			next if ($c2->{weight} == 0);
			my $c = new RelaxGraph::Constraint2;
			&RelaxGraphc::Constraint2_weight_set( $c, $c2->{weight});
			foreach my $vei (@{$c2->{adj}}) {
				&RelaxGraph::PushAdjC2( $c, $vei);
			}
			
			# influence rule
			for (my $ii = 0; $ii < @{$c2->{inf}}; $ii++) {
				for (my $jj = 0; $jj < @{$c2->{inf}[$ii]}; $jj++) {
					&RelaxGraph::PushInfC2( $c, $ii, $c2->{inf}[$ii][$jj]);
				}
			}
			
			&RelaxGraph::PushConstraint2( $n, $c);
		}
		
		$tnodes++;
		RelaxGraphc::PushNode( $problem, $n);
	}
	
	RelaxGraphc::Relaxation( $problem);
	
	for (my $i = 0; $i < $tnodes; $i++) {
		$vertices->[$i]->defLabel(&RelaxGraph::ReadNodeLabel( $problem, $i));
	}
	
	$obj->GenResponse($doc, $vertices);
	
	# free memory
	$problem = undef;
	$vertices = undef;
	$edges = undef;
	$graph->Clear;
}


sub GenResponse
{
	my ($obj, $doc, $vertices) = @_;
	my @response;
	
	foreach my $v (@$vertices) {
		my $m = $v->mention;
		$m->{chain} = $v->defLabel;
		push(@response, $m);
	}
	
	$doc->SaveLevel(\@response, 'response');
}

1;
