package Vertex;
use strict;
use Mention;
use Edge;

sub new {
	my ($pkg, $m, $e) = @_;
	my $v = bless { _mention => $m, _edges => $e }, $pkg;
	$v->id($m->index) if ($m);
	return $v;
}

sub id {
	my $obj = shift;
	@_ ? $obj->{id} = shift : $obj->{id};
}

sub graphid {
	my $obj = shift;
	@_ ? $obj->{graphid} = shift : $obj->{graphid};
}

sub mention {
	my $obj = shift;
	@_ ? $obj->{_mention} = shift : $obj->{_mention};
}

sub edges {
	my $obj = shift;
	@_ ? $obj->{_edges} = shift : $obj->{_edges};
}

sub nLabels {
	my ($obj, $n) = @_;
	if ($n) {
		$obj->{_nLabels} = $n;
		@{$obj->{_label}} = ();
		push ( @{$obj->{_label}}, 0) for (1..$n);
		@{$obj->{_nextLabel}} = ();
		push ( @{$obj->{_nextLabel}}, 0) for (1..$n);
	}
	else {
		return $obj->{_nLabels};
	}
}

sub label {
	my $obj = shift;
	@_ ? $obj->{_label} = shift : $obj->{_label};
}

sub nextLabel {
	my $obj = shift;
	@_ ? $obj->{_nextLabel} = shift : $obj->{_nextLabel};
}

sub UpdateLabels {
	my $obj = shift;
	@{$obj->{_label}} = @{$obj->{_nextLabel}};
}

# label definitiu (si no existeix, retorna maxLabel
sub defLabel {
	my $obj = shift;
	if (@_) {
		$obj->{_defLabel} = shift;
	}
	elsif (defined($obj->{_defLabel})) {
		return $obj->{_defLabel};
	}
	
	return $obj->maxLabel;
}

sub maxLabel {
	my $obj = shift;
	my $max = undef;
	my $ml = 0;
	for (my $l = 0; $l < $obj->{_nLabels}; $l++) {
		if (!defined($max) || $obj->{_label}->[$l] > $max) {
			$max = $obj->{_label}->[$l];
			$ml = $l;
		}
	}
	
	if (wantarray) {
		return ($ml, $max);
	}
	return $ml;
}

sub adjacencies
{
	my $obj = shift;
	
	if (!defined($obj->{_adjacencies})) {
		@{$obj->{_adjacencies}} = ();
		foreach my $edge (@{$obj->{_edges}}) {
			if ($edge->incident($obj)) {
				push(@{$obj->{_adjacencies}}, $edge);
			}
		}
	}
	
	return $obj->{_adjacencies};
}

sub isAdjacent
{
	my ($obj, $other) = @_;
	
	foreach my $e (@{$obj->adjacencies}) {
		if ($e->incident($other)) {
			return 1;
		}
	}
	return 0;
}

sub c2 {
	my $obj = shift;
	if (!defined($obj->{_c2})) {
		@{$obj->{_c2}} = ();
	}
	@_ ? $obj->{_c2} = shift : $obj->{_c2};
}

sub DESTROY {
	my $obj = shift;
	undef($obj->{_adjacencies});
	undef($obj->{_mention});
	undef($obj->{_edges});
	undef($obj->{_c2});
}

1;
