#!/usr/bin/perl

use FindBin qw( $RealBin );
use lib "$RealBin/lib";


use strict;
use Getopt::Long;
use RelaxCor;
use CorScorer;
use POSIX ":sys_wait_h"; # fork

###############################################################################
# globals

use Globals qw ($PATH_HOME $PATH_LOGS $STEM);

my $VERBOSE = 0;
my $trueMentions = 0;

my $DEFAULT_SECTION = 'sec-default';
my $DEFAULT_MODEL = 'default';

my $PARALEL = 2;

my $KEY_FILE_COLS = undef;

###############################################################################
# functions

sub Solve
{
	my ($modelAlias, $section, $files, $options) = @_;
	my $c = 0;
	my $kids = 0;
	foreach my $file (@$files) {
		$c++;
		print "($c / " . scalar(@$files) . ') ';
		
		my $n = $options->{Nprune};
		my $command = "$RealBin/run-relaxcor.pl --model=$modelAlias --section=$section --N=$n";
		if (defined($options->{emodel})) {
			$command .= " --emodel=$options->{emodel}";
		}
		$command .= " s $file";
		
		if (!defined($PARALEL) || $PARALEL < 2) {
			system($command);
		}
		else {
			my $pid = fork;
			if ($pid == 0) { #child
				system($command);
				exit(0);
			}
			$kids++;
			
			# recull zombies
			my $child;
			do {
				$child = waitpid(-1, WNOHANG);
				$kids-- if ($child > 0);
			} while ($child > 0);
			
			# Si segeuix havent-hi 8 o mes fills executant-se esperem
			# que acabi el primer
			if ($kids >= $PARALEL) {
				wait;
				$kids--;
			}
		}
	}

	if (defined($PARALEL)) {
		while ($kids > 0) {
			wait;
			$kids--;
		}
	}
}

sub CalculateGrid
{
	my ($modelAlias, $section, $files, $keyfile, $opt) = @_;
	
	my $modelFile = "$PATH_HOME/models/$modelAlias/$STEM-$section";
	
	my $model = new Model($modelFile);
	$model->alias($modelAlias);
	$model->verbosity($VERBOSE);
	$model->trueMentions($trueMentions);

	for (my $Nprune = 2; $Nprune <= 10; $Nprune += 2) {
		print "Nprune: $Nprune\n";

		my $inib = 1;
		for (my $bi = $inib; $bi < 7; $bi++) {
			my $balance = '0.' . $bi;
			print "Balance: $balance\n";
			
			# Recalculem pesos en funcio del balance
			#my %options = ( balance => $balance, minAplica => 5, Nprune => $Nprune);
			$opt->{balance} = $balance;
			$opt->{minAplica} = 5;
			$opt->{Nprune} = $Nprune;
			
			&RelaxCor::WeightConstraints($model, $opt);
			
			my $logFile = "$PATH_LOGS/balance-$modelAlias-$section.txt";
			if (defined($opt->{emodel})) {
				$logFile = "$PATH_LOGS/balance-$modelAlias-$opt->{emodel}-$section.txt";
			}
			my $logFileOutTmp = $logFile . '-' . $Nprune . '-' . $balance . '.out';
			
			# Comentar si es volen matxacar els que ja estaven fets
			next if (-e $logFileOutTmp);
			
			# Resol
			Solve($modelAlias, $section, $files, $opt);
			
			
			if (defined($keyfile)) {
				system("rm $logFileOutTmp") if (-e $logFileOutTmp);
				foreach my $file (@$files) {
					my $netfile = $file;
					$netfile =~ s/^.*\/([^\/]*?)$/$1/;
					system("echo \'#begin document $netfile\' >> $logFileOutTmp");
					system("cat $file >> $logFileOutTmp");
					system("echo \'#end document $netfile\' >> $logFileOutTmp");
				}
			}
			
			
			open (LOG, ">>$logFile");
			my @metriques = ('ceafm', 'ceafe', 'muc', 'bcub');
			foreach my $metric (@metriques) {
				my ($acumNR, $acumDR, $acumNP, $acumDP) = (0, 0, 0, 0);
				if (!defined($keyfile)) {
					# We assume that "$file.gold" exists for each $file
					foreach my $file (@$files) {
						my ($dNR, $dDR, $dNP, $dDP) = &CorScorer::Score( $metric, "$file.gold", "$file.out", '__single');
						$acumNR += $dNR;
						$acumDR += $dDR;
						$acumNP += $dNP;
						$acumDP += $dDP;
					}
				}
				else {
					($acumNR, $acumDR, $acumNP, $acumDP) = &CorScorer::Score( $metric,
																		$keyfile, $logFileOutTmp );
				}
				my $p = ($acumDP) ? $acumNP / $acumDP : 0;
				my $r = ($acumDR) ? $acumNR / $acumDR : 0;
				my $f1 = ($p + $r) ? 2 * $p * $r / ($p + $r) : 0;
				print "$metric ==>\tPRE: $p, REC: $r, F1: $f1\n";
				print LOG "N: $Nprune | balance: $balance | $metric | PRE: $p, REC: $r, F1: $f1\n";
			}
			close (LOG);
		}
	}
}

sub FindMax
{
	my ($metrica, $file) = @_;

	my @l;
	if (!defined($metrica) || $metrica eq '') {
		$metrica = 'muc';
	}
	print "Metrica: $metrica\n";
	if ($metrica eq 'm') {
		open (F, $file) || die "Can not open $file: $!";
		@l = <F>;
	}
	else {
		@l = `grep '$metrica' $file`;
	}

	chomp($l[0]);
	my ($mRec, $mPre) = (undef, undef);
	my ($max, $Nmax, $balanceMax) = (-1,0,0,);
	my $jj = 0;
	my $f1acum = 0;
	if ($l[0] =~ /F1\:\s*?(0\.\d+?)\b/) {
		foreach my $line (@l) {
			chomp($line);
			my ($N) = ($line =~ /N\: (\d+) /);
			my $balance;
			if ($line =~ /balance\: (0?[\.\d]+) /) {
				$balance = $1;
			}
			
			my ($pre, $rec, $f1) = ($line =~ /PRE:\s*?(0\.\d+?), REC:\s*?(0\.\d+?), F1\:\s*?(0\.\d+?)\b/);
			
			if ($metrica eq 'm') {
				if ($line =~ /(ceafe|muc|bcub)/) {
					$f1acum += $f1;
					$jj++;
					if ($jj == 3) {
						$f1 = $f1acum / 3;
						$jj = 0;
						$f1acum = 0;
						if ($f1 > $max) {
							$max = $f1;
							$Nmax = $N;
							$balanceMax = $balance;
						}
					}
				}
			}
			else {
				if ($f1 > $max) {
					$max = $f1;
					$Nmax = $N;
					$balanceMax = $balance;
					$mRec = $rec;
					$mPre = $pre;
				}
			}
		}
	}

	print "Max: $max => N: $Nmax, balance: $balanceMax\n";
	if (defined($mRec)) {
		print "Recall: $mRec\tPrecisio: $mPre\n";
	}
}

###############################################################################
# main

my $result;
my %opt;
if (@ARGV) {
	$result = GetOptions (\%opt, 'parallel=i', 'keyfile=s', 'measure=s', 'findmax', 'trueMentions', 'emodel=s');
}

if (!$result || @ARGV < 3) {
	print "\n\tus: ./run-devel.pl [options] <model> <section> <files> ...\n\n";
	exit;
}

my $modelAlias = shift(@ARGV);
my $section = shift(@ARGV);
my @files = @ARGV;

my $keyfile = undef;
if (defined($opt{keyfile})) {
	$keyfile = $opt{keyfile};
	if ($keyfile eq '1') {
		$keyfile = $KEY_FILE_COLS;
	}
}

if (defined($opt{parallel})) {
	$PARALEL = $opt{parallel};
}

$trueMentions = defined($opt{trueMentions}) ? 1 : 0;

print "Model: $modelAlias\n";
if ($keyfile) {
	print "keyfile: $keyfile\n";
}

if (!defined($opt{findmax})) {
	CalculateGrid($modelAlias, $section, \@files, $keyfile, \%opt);
}

my $measure = defined($opt{measure}) ? $opt{measure} : 'muc';

my $logFile = "$PATH_LOGS/balance-$modelAlias-$section.txt";
if (defined($opt{emodel})) {
	$logFile = "$PATH_LOGS/balance-$modelAlias-$opt{emodel}-$section.txt";
}
FindMax($measure, $logFile);
