#!/usr/bin/perl

# Mostra info sobre les mencions dels documents

use FindBin qw( $RealBin );
use lib "$RealBin/lib";

use strict;
use Getopt::Long;
use Document;
use Model;

###############################################################################
# funcions


###############################################################################
# Arguments d'entrada

my %opt = ( mention => -1 );

my $result = 0;
my $file;
my ($mi, $mj);

if (@ARGV < 3) {
	print '
	use: ./show-features.pl <m_i> <m_j> <document> 

';
	exit;
}

###############################################################################
# main

$file = pop(@ARGV);
if (!-e $file) {
	die "Document $file does not exist.";
}

my $doc = new Document($file);
my $mentions = $doc->GetMentions(1, 1);

$mi = shift(@ARGV);
$mj = shift(@ARGV);

if ($mi !~ /^\d+$/ || $mj !~ /^\d+$/) {
	die "m_i and m_j must be numbers";
}

if ($mi < 0 || $mi >= @$mentions) {
	die "m_i out of range (document has " . scalar(@$mentions) . " mentions)";
}

if ($mj < 0 || $mj >= @$mentions) {
	die "m_i out of range (document has " . scalar(@$mentions) . " mentions)";
}

print "$mi: " . $mentions->[$mi]->string . ' [' . $mentions->[$mi]->head_str . "]\n";
print "$mj: " . $mentions->[$mj]->string . ' [' . $mentions->[$mj]->head_str . "]\n";

my $model = new Model('any');
my $coref = $doc->Corefer($mentions->[$mi], $mentions->[$mj]) || 0;
print "Coreferent: $coref\n";

my @features = $model->features;
my @values = $model->GenPairFeatures($mentions->[$mi], $mentions->[$mj]);

for (my $i = 0; $i < @features; $i++) {
	print "$features[$i]: $values[$i]\n";
}

